import numpy as np

# from previous exercise
from finite_difference import finite_difference
# already provided in this exercise
from Phi import Phi
from Phi_casadi import derive_Phi_casadi
# your implementation in this exercise
from i_trick import i_trick
from Phi_FAD import Phi_FAD
from Phi_BAD import Phi_BAD


param = {
    'N': 200,    # number of discretization steps
    'x0':  2,    # initial condition
    'T':  5,    # terminal time
    'q': 20,    # terminal weight
}
h = param['T'] / param['N']

Utst = np.random.rand(param['N'])  # random control trajectory

# finite differences on nonlinear part
# TODO: get finite_difference() from previous exercise
[F1, J1] = finite_difference(Phi, Utst, param)

# imaginary trick 
# TODO: implement imaginary trick in i_trick
[F2, J2] = i_trick(Phi, Utst, param)

# forward AD
# TODO: implement forward mode of AD in Phi_FAD
[F3, J3] = Phi_FAD(Utst, param)

# backward AD
# TODO: implement backward mode of AD in Phi_BAD
[F4, J4] = Phi_BAD(Utst, param)

# running this script adds the casadi function Phi_casadi(U) to workspace
# TODO: complete Phi_casadi.py
Phi_casadi = derive_Phi_casadi(param)
# evaluate the casadi function
[F5, J5] = Phi_casadi(Utst)


# check results
print("Error between imaginary trick and finite differences:")
print(np.amax(np.absolute(J2-J1)))

print("Error between imaginary trick and forward AD:")
print(np.amax(np.absolute(J2-J3)))

print("Error between imaginary trick and backward AD:")
print(np.amax(np.absolute(J2-J4)))

print("Error between imaginary trick and casadi:")
print(np.amax(np.absolute(J2-J5)))
